"""Tests for `$ fio cat`."""


from click.testing import CliRunner

from fiona.fio.main import main_group


def test_one(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(main_group, ['cat', path_coutwildrnp_shp])
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 67


def test_two(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(main_group, ['cat', path_coutwildrnp_shp, path_coutwildrnp_shp])
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 134


def test_bbox_no(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ['cat', path_coutwildrnp_shp, '--bbox', '0,10,80,20'],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output == ""


def test_bbox_yes(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ['cat', path_coutwildrnp_shp, '--bbox', '-109,37,-107,39'],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 19


def test_bbox_yes_two_files(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ['cat', path_coutwildrnp_shp, path_coutwildrnp_shp, '--bbox', '-109,37,-107,39'],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 38


def test_bbox_json_yes(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ['cat', path_coutwildrnp_shp, '--bbox', '[-109,37,-107,39]'],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 19


def test_bbox_where(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ['cat', path_coutwildrnp_shp, '--bbox', '-120,40,-100,50',
         '--where', "NAME LIKE 'Mount%'"],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 4


def test_where_no(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ['cat', path_coutwildrnp_shp, '--where', "STATE LIKE '%foo%'"],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output == ""


def test_where_yes(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ['cat', path_coutwildrnp_shp, '--where', "NAME LIKE 'Mount%'"],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 9


def test_where_yes_two_files(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(
        main_group,
        ['cat', path_coutwildrnp_shp, path_coutwildrnp_shp,
         '--where', "NAME LIKE 'Mount%'"],
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 18


def test_where_fail(data_dir):
    runner = CliRunner()
    result = runner.invoke(main_group, ['cat', '--where', "NAME=3",
                           data_dir])
    assert result.exit_code != 0


def test_multi_layer(data_dir):
    layerdef = "1:coutwildrnp,1:coutwildrnp"
    runner = CliRunner()
    result = runner.invoke(
        main_group, ['cat', '--layer', layerdef, data_dir])
    assert result.output.count('"Feature"') == 134


def test_multi_layer_fail(data_dir):
    runner = CliRunner()
    result = runner.invoke(main_group, ['cat', '--layer', '200000:coutlildrnp',
                           data_dir])
    assert result.exit_code != 0


def test_vfs(path_coutwildrnp_zip):
    runner = CliRunner()
    result = runner.invoke(main_group, [
        'cat', f'zip://{path_coutwildrnp_zip}'])
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 67


def test_dst_crs_epsg3857(path_coutwildrnp_shp):
    """Confirm fix of issue #952"""
    runner = CliRunner()
    result = runner.invoke(
        main_group, ["cat", "--dst-crs", "EPSG:3857", path_coutwildrnp_shp]
    )
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 67



"""Tests for `$ fio collect`."""


import json
import sys

from click.testing import CliRunner
import pytest

# from fiona.fio import collect
from fiona.fio.main import main_group


def test_collect_rs(feature_seq_pp_rs):
    runner = CliRunner()
    result = runner.invoke(
        main_group, ['collect', '--src-crs', 'EPSG:3857'],
        feature_seq_pp_rs,
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 2


def test_collect_no_rs(feature_seq):
    runner = CliRunner()
    result = runner.invoke(
        main_group, ['collect', '--src-crs', 'EPSG:3857'],
        feature_seq,
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 2


def test_collect_ld(feature_seq):
    runner = CliRunner()
    result = runner.invoke(
        main_group, ['collect', '--with-ld-context', '--add-ld-context-item', 'foo=bar'],
        feature_seq,
        catch_exceptions=False)
    assert result.exit_code == 0
    assert '"@context": {' in result.output
    assert '"foo": "bar"' in result.output


def test_collect_rec_buffered(feature_seq):
    runner = CliRunner()
    result = runner.invoke(main_group, ['collect', '--record-buffered'], feature_seq)
    assert result.exit_code == 0
    assert '"FeatureCollection"' in result.output


def test_collect_noparse(feature_seq):
    runner = CliRunner()
    result = runner.invoke(
        main_group, ['collect', '--no-parse'],
        feature_seq,
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 2
    assert len(json.loads(result.output)['features']) == 2


def test_collect_noparse_records(feature_seq):
    runner = CliRunner()
    result = runner.invoke(
        main_group, ['collect', '--no-parse', '--record-buffered'],
        feature_seq,
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 2
    assert len(json.loads(result.output)['features']) == 2


def test_collect_src_crs(feature_seq):
    runner = CliRunner()
    result = runner.invoke(
        main_group, ['collect', '--no-parse', '--src-crs', 'epsg:4326'],
        feature_seq,
        catch_exceptions=False)
    assert result.exit_code == 2


def test_collect_noparse_rs(feature_seq_pp_rs):
    runner = CliRunner()
    result = runner.invoke(
        main_group, ['collect', '--no-parse'],
        feature_seq_pp_rs,
        catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 2
    assert len(json.loads(result.output)['features']) == 2



import pytest
from fiona._env import get_gdal_version_tuple

import fiona
from fiona.drvsupport import supported_drivers, _driver_supports_mode
from fiona.errors import DriverError
from fiona.env import GDALVersion
from tests.conftest import get_temp_filename


def test_bounds_point():
    g = {"type": "Point", "coordinates": [10, 10]}
    assert fiona.bounds(g) == (10, 10, 10, 10)


def test_bounds_line():
    g = {"type": "LineString", "coordinates": [[0, 0], [10, 10]]}
    assert fiona.bounds(g) == (0, 0, 10, 10)


def test_bounds_polygon():
    g = {"type": "Polygon", "coordinates": [[[0, 0], [10, 10], [10, 0]]]}
    assert fiona.bounds(g) == (0, 0, 10, 10)


def test_bounds_z():
    g = {"type": "Point", "coordinates": [10, 10, 10]}
    assert fiona.bounds(g) == (10, 10, 10, 10)


# MapInfo File driver requires that the bounds (geographical extents) of a new file
# be set before writing the first feature (https://gdal.org/drivers/vector/mitab.html)


@pytest.mark.parametrize(
    "driver",
    [
        driver
        for driver in supported_drivers
        if _driver_supports_mode(driver, "w") and not driver == "MapInfo File"
    ],
)
def test_bounds(tmpdir, driver, testdata_generator):
    """Test if bounds are correctly calculated after writing."""
    if driver == "BNA" and GDALVersion.runtime() < GDALVersion(2, 0):
        pytest.skip("BNA driver segfaults with gdal 1.11")
    if driver == "ESRI Shapefile" and get_gdal_version_tuple() < (3, 1):
        pytest.skip(
            "Bug in GDALs Shapefile driver: https://github.com/OSGeo/gdal/issues/2269"
        )

    range1 = list(range(0, 5))
    range2 = list(range(5, 10))
    schema, crs, records1, records2, test_equal = testdata_generator(
        driver, range1, range2
    )

    if not schema["geometry"] == "Point":
        pytest.skip("Driver does not support point geometries")

    filename = get_temp_filename(driver)
    path = str(tmpdir.join(filename))

    def calc_bounds(records):
        xs = []
        ys = []
        for r in records:
            xs.append(r.geometry["coordinates"][0])
            ys.append(r.geometry["coordinates"][1])
        return min(xs), max(xs), min(ys), max(ys)

    with fiona.open(path, "w", driver=driver, schema=schema) as c:
        c.writerecords(records1)

        try:
            bounds = c.bounds
            assert bounds == calc_bounds(records1)
        except Exception as e:
            assert isinstance(e, DriverError)

        c.writerecords(records2)

        try:
            bounds = c.bounds
            assert bounds == calc_bounds(records1 + records2)
        except Exception as e:
            assert isinstance(e, DriverError)



"""Tests of the ogrext.Session class"""
import pytest

import fiona
from fiona.errors import GDALVersionError, UnsupportedOperation
from .conftest import gdal_version


def test_get(path_coutwildrnp_shp):
    with fiona.open(path_coutwildrnp_shp) as col:
        feat3 = col.get(2)
        assert feat3['properties']['NAME'] == 'Mount Zirkel Wilderness'


@pytest.mark.parametrize("layer, namespace, tags", [
    (None, None, {"test_tag1": "test_value1", "test_tag2": "test_value2"}),
    (None, "test", {"test_tag1": "test_value1", "test_tag2": "test_value2"}),
    (None, None, {}),
    (None, "test", {}),  
    ("layer", None, {"test_tag1": "test_value1", "test_tag2": "test_value2"}),
    ("layer", "test", {"test_tag1": "test_value1", "test_tag2": "test_value2"}),
    ("layer", None, {}),
    ("layer", "test", {}),  
])
@pytest.mark.skipif(gdal_version.major < 2, reason="Broken on GDAL 1.x")
def test_update_tags(layer, namespace, tags, tmpdir):
    test_geopackage = str(tmpdir.join("test.gpkg"))
    schema = {'properties': {'CDATA1': 'str:254'}, 'geometry': 'Polygon'}
    with fiona.Env(), fiona.open(
            test_geopackage, "w", driver="GPKG", schema=schema, layer=layer) as gpkg:
        assert gpkg.tags() == {}
        gpkg.update_tags(tags, ns=namespace)

    with fiona.Env(), fiona.open(test_geopackage, layer=layer) as gpkg:
        assert gpkg.tags(ns=namespace) == tags
        if namespace is not None:
            assert gpkg.tags() == {}
        with pytest.raises(UnsupportedOperation):
            gpkg.update_tags({}, ns=namespace)


@pytest.mark.parametrize("layer, namespace", [
    (None, None),
    (None, "test"),
    ("test", None),
    ("test", "test"),
])
@pytest.mark.skipif(gdal_version.major < 2, reason="Broken on GDAL 1.x")
def test_update_tag_item(layer, namespace, tmpdir):
    test_geopackage = str(tmpdir.join("test.gpkg"))
    schema = {'properties': {'CDATA1': 'str:254'}, 'geometry': 'Polygon'}
    with fiona.Env(), fiona.open(
            test_geopackage, "w", driver="GPKG", schema=schema, layer=layer) as gpkg:
        assert gpkg.get_tag_item("test_tag1", ns=namespace) is None
        gpkg.update_tag_item("test_tag1", "test_value1", ns=namespace)

    with fiona.Env(), fiona.open(test_geopackage, layer=layer) as gpkg:
        if namespace is not None:
            assert gpkg.get_tag_item("test_tag1") is None
        assert gpkg.get_tag_item("test_tag1", ns=namespace) == "test_value1"
        with pytest.raises(UnsupportedOperation):
            gpkg.update_tag_item("test_tag1", "test_value1", ns=namespace)


@pytest.mark.skipif(gdal_version.major >= 2, reason="Only raises on GDAL 1.x")
def test_gdal_version_error(tmpdir):
    test_geopackage = str(tmpdir.join("test.gpkg"))
    schema = {'properties': {'CDATA1': 'str:254'}, 'geometry': 'Polygon'}
    with fiona.Env(), fiona.open(
            test_geopackage, "w", driver="GPKG", schema=schema, layer="layer") as gpkg:
        with pytest.raises(GDALVersionError):
            gpkg.update_tags({"test_tag1": "test_value1"}, ns="test")
        with pytest.raises(GDALVersionError):
            gpkg.update_tag_item("test_tag1", "test_value1", ns="test")
        with pytest.raises(GDALVersionError):
            gpkg.tags()
        with pytest.raises(GDALVersionError):
            gpkg.get_tag_item("test_tag1")



# Testing collections and workspaces

from collections import OrderedDict
import datetime
import logging
import random
import re
import sys

import pytest

import fiona
from fiona.collection import Collection
from fiona.drvsupport import supported_drivers, driver_mode_mingdal
from fiona.env import getenv, GDALVersion
from fiona.errors import (
    AttributeFilterError,
    FionaValueError,
    DriverError,
    FionaDeprecationWarning,
)
from fiona.model import Feature, Geometry

from .conftest import WGS84PATTERN, get_temp_filename


class TestSupportedDrivers:
    def test_shapefile(self):
        assert "ESRI Shapefile" in supported_drivers
        assert set(supported_drivers["ESRI Shapefile"]) == set("raw")

    def test_map(self):
        assert "MapInfo File" in supported_drivers
        assert set(supported_drivers["MapInfo File"]) == set("raw")


class TestCollectionArgs:
    def test_path(self):
        with pytest.raises(TypeError):
            Collection(0)

    def test_mode(self):
        with pytest.raises(TypeError):
            Collection("foo", mode=0)

    def test_driver(self):
        with pytest.raises(TypeError):
            Collection("foo", mode="w", driver=1)

    def test_schema(self):
        with pytest.raises(TypeError):
            Collection("foo", mode="w", driver="ESRI Shapefile", schema=1)

    def test_crs(self):
        with pytest.raises(TypeError):
            Collection("foo", mode="w", driver="ESRI Shapefile", schema=0, crs=1)

    def test_encoding(self):
        with pytest.raises(TypeError):
            Collection("foo", mode="r", encoding=1)

    def test_layer(self):
        with pytest.raises(TypeError):
            Collection("foo", mode="r", layer=0.5)

    def test_vsi(self):
        with pytest.raises(TypeError):
            Collection("foo", mode="r", vsi="git")

    def test_archive(self):
        with pytest.raises(TypeError):
            Collection("foo", mode="r", archive=1)

    def test_write_numeric_layer(self):
        with pytest.raises(ValueError):
            Collection("foo", mode="w", layer=1)

    def test_write_geojson_layer(self):
        with pytest.raises(ValueError):
            Collection("foo", mode="w", driver="GeoJSON", layer="foo")

    def test_append_geojson(self):
        with pytest.raises(ValueError):
            Collection("foo", mode="w", driver="ARCGEN")


class TestOpenException:
    def test_no_archive(self):
        with pytest.warns(FionaDeprecationWarning), pytest.raises(DriverError):
            fiona.open("/", mode="r", vfs="zip:///foo.zip")


class TestReading:
    @pytest.fixture(autouse=True)
    def shapefile(self, path_coutwildrnp_shp):
        self.c = fiona.open(path_coutwildrnp_shp, "r")
        yield
        self.c.close()

    def test_open_repr(self, path_coutwildrnp_shp):
        assert repr(self.c) == (
            "<open Collection '{path}:coutwildrnp', mode 'r' "
            "at {hexid}>".format(hexid=hex(id(self.c)), path=path_coutwildrnp_shp)
        )

    def test_closed_repr(self, path_coutwildrnp_shp):
        self.c.close()
        assert repr(self.c) == (
            "<closed Collection '{path}:coutwildrnp', mode 'r' "
            "at {hexid}>".format(hexid=hex(id(self.c)), path=path_coutwildrnp_shp)
        )

    def test_path(self, path_coutwildrnp_shp):
        assert self.c.path == path_coutwildrnp_shp

    def test_name(self):
        assert self.c.name == "coutwildrnp"

    def test_mode(self):
        assert self.c.mode == "r"

    def test_encoding(self):
        assert self.c.encoding is None

    def test_iter(self):
        assert iter(self.c)

    def test_closed_no_iter(self):
        self.c.close()
        with pytest.raises(ValueError):
            iter(self.c)

    def test_len(self):
        assert len(self.c) == 67

    def test_closed_len(self):
        # Len is lazy, it's never computed in this case. TODO?
        self.c.close()
        assert len(self.c) == 0

    def test_len_closed_len(self):
        # Lazy len is computed in this case and sticks.
        len(self.c)
        self.c.close()
        assert len(self.c) == 67

    def test_driver(self):
        assert self.c.driver == "ESRI Shapefile"

    def test_closed_driver(self):
        self.c.close()
        assert self.c.driver is None

    def test_driver_closed_driver(self):
        self.c.driver
        self.c.close()
        assert self.c.driver == "ESRI Shapefile"

    def test_schema(self):
        s = self.c.schema["properties"]
        assert s["PERIMETER"] == "float:24.15"
        assert s["NAME"] == "str:80"
        assert s["URL"] == "str:101"
        assert s["STATE_FIPS"] == "str:80"
        assert s["WILDRNP020"] == "int:10"

    def test_closed_schema(self):
        # Schema is lazy too, never computed in this case. TODO?
        self.c.close()
        assert self.c.schema is None

    def test_schema_closed_schema(self):
        self.c.schema
        self.c.close()
        assert sorted(self.c.schema.keys()) == ["geometry", "properties"]

    def test_crs(self):
        crs = self.c.crs
        assert crs["init"] == "epsg:4326"

    def test_crs_wkt(self):
        crs = self.c.crs_wkt
        assert re.match(WGS84PATTERN, crs)

    def test_closed_crs(self):
        # Crs is lazy too, never computed in this case. TODO?
        self.c.close()
        assert self.c.crs is None

    def test_crs_closed_crs(self):
        self.c.crs
        self.c.close()
        assert sorted(self.c.crs.keys()) == ["init"]

    def test_meta(self):
        assert sorted(self.c.meta.keys()) == ["crs", "crs_wkt", "driver", "schema"]

    def test_profile(self):
        assert sorted(self.c.profile.keys()) == ["crs", "crs_wkt", "driver", "schema"]

    def test_bounds(self):
        assert self.c.bounds[0] == pytest.approx(-113.564247)
        assert self.c.bounds[1] == pytest.approx(37.068981)
        assert self.c.bounds[2] == pytest.approx(-104.970871)
        assert self.c.bounds[3] == pytest.approx(41.996277)

    def test_context(self, path_coutwildrnp_shp):
        with fiona.open(path_coutwildrnp_shp, "r") as c:
            assert c.name == "coutwildrnp"
            assert len(c) == 67
            assert c.crs
        assert c.closed

    def test_iter_one(self):
        itr = iter(self.c)
        f = next(itr)
        assert f.id == "0"
        assert f.properties["STATE"] == "UT"

    def test_iter_list(self):
        f = list(self.c)[0]
        assert f.id == "0"
        assert f.properties["STATE"] == "UT"

    def test_re_iter_list(self):
        f = list(self.c)[0]  # Run through iterator
        f = list(self.c)[0]  # Run through a new, reset iterator
        assert f.id == "0"
        assert f.properties["STATE"] == "UT"

    def test_getitem_one(self):
        f = self.c[0]
        assert f.id == "0"
        assert f.properties["STATE"] == "UT"

    def test_getitem_iter_combo(self):
        i = iter(self.c)
        f = next(i)
        f = next(i)
        assert f.id == "1"
        f = self.c[0]
        assert f.id == "0"
        f = next(i)
        assert f.id == "2"

    def test_no_write(self):
        with pytest.raises(OSError):
            self.c.write({})

    def test_iter_items_list(self):
        i, f = list(self.c.items())[0]
        assert i == 0
        assert f.id == "0"
        assert f.properties["STATE"] == "UT"

    def test_iter_keys_list(self):
        i = list(self.c.keys())[0]
        assert i == 0

    def test_in_keys(self):
        assert 0 in self.c.keys()
        assert 0 in self.c


class TestReadingPathTest:
    def test_open_path(self, path_coutwildrnp_shp):
        pathlib = pytest.importorskip("pathlib")
        with fiona.open(pathlib.Path(path_coutwildrnp_shp)) as collection:
            assert collection.name == "coutwildrnp"


@pytest.mark.usefixtures("unittest_path_coutwildrnp_shp")
class TestIgnoreFieldsAndGeometry:
    def test_without_ignore(self):
        with fiona.open(self.path_coutwildrnp_shp, "r") as collection:
            assert "AREA" in collection.schema["properties"].keys()
            assert "STATE" in collection.schema["properties"].keys()
            assert "NAME" in collection.schema["properties"].keys()
            assert "geometry" in collection.schema.keys()

            feature = next(iter(collection))
            assert feature["properties"]["AREA"] is not None
            assert feature["properties"]["STATE"] is not None
            assert feature["properties"]["NAME"] is not None
            assert feature["geometry"] is not None

    def test_ignore_fields(self):
        with fiona.open(
            self.path_coutwildrnp_shp, "r", ignore_fields=["AREA", "STATE"]
        ) as collection:
            assert "AREA" not in collection.schema["properties"].keys()
            assert "STATE" not in collection.schema["properties"].keys()
            assert "NAME" in collection.schema["properties"].keys()
            assert "geometry" in collection.schema.keys()

            feature = next(iter(collection))
            assert "AREA" not in feature["properties"].keys()
            assert "STATE" not in feature["properties"].keys()
            assert feature["properties"]["NAME"] is not None
            assert feature["geometry"] is not None

    def test_ignore_invalid_field_missing(self):
        with fiona.open(
            self.path_coutwildrnp_shp, "r", ignore_fields=["DOES_NOT_EXIST"]
        ):
            pass

    def test_ignore_invalid_field_not_string(self):
        with pytest.raises(TypeError):
            with fiona.open(self.path_coutwildrnp_shp, "r", ignore_fields=[42]):
                pass

    def test_include_fields(self):
        with fiona.open(
            self.path_coutwildrnp_shp, "r", include_fields=["AREA", "STATE"]
        ) as collection:
            assert sorted(collection.schema["properties"]) == ["AREA", "STATE"]
            assert "geometry" in collection.schema.keys()

            feature = next(iter(collection))
            assert sorted(feature["properties"]) == ["AREA", "STATE"]
            assert feature["properties"]["AREA"] is not None
            assert feature["properties"]["STATE"] is not None
            assert feature["geometry"] is not None

    def test_include_fields__geom_only(self):
        with fiona.open(
            self.path_coutwildrnp_shp, "r", include_fields=()
        ) as collection:
            assert sorted(collection.schema["properties"]) == []
            assert "geometry" in collection.schema.keys()

            feature = next(iter(collection))
            assert sorted(feature["properties"]) == []
            assert feature["geometry"] is not None

    def test_include_fields__ignore_fields_error(self):
        with pytest.raises(ValueError):
            with fiona.open(
                self.path_coutwildrnp_shp,
                "r",
                include_fields=["AREA"],
                ignore_fields=["STATE"],
            ) as collection:
                pass

    def test_ignore_geometry(self):
        with fiona.open(
            self.path_coutwildrnp_shp, "r", ignore_geometry=True
        ) as collection:
            assert "AREA" in collection.schema["properties"].keys()
            assert "STATE" in collection.schema["properties"].keys()
            assert "NAME" in collection.schema["properties"].keys()
            assert "geometry" not in collection.schema.keys()

            feature = next(iter(collection))
            assert feature.properties["AREA"] is not None
            assert feature.properties["STATE"] is not None
            assert feature.properties["NAME"] is not None
            assert feature.geometry is None


class TestFilterReading:
    @pytest.fixture(autouse=True)
    def shapefile(self, path_coutwildrnp_shp):
        self.c = fiona.open(path_coutwildrnp_shp, "r")
        yield
        self.c.close()

    def test_filter_1(self):
        results = list(self.c.filter(bbox=(-120.0, 30.0, -100.0, 50.0)))
        assert len(results) == 67
        f = results[0]
        assert f.id == "0"
        assert f.properties["STATE"] == "UT"

    def test_filter_reset(self):
        results = list(self.c.filter(bbox=(-112.0, 38.0, -106.0, 40.0)))
        assert len(results) == 26
        results = list(self.c.filter())
        assert len(results) == 67

    def test_filter_mask(self):
        mask = Geometry.from_dict(
            **{
                "type": "Polygon",
                "coordinates": (
                    ((-112, 38), (-112, 40), (-106, 40), (-106, 38), (-112, 38)),
                ),
            }
        )
        results = list(self.c.filter(mask=mask))
        assert len(results) == 26

    def test_filter_where(self):
        results = list(self.c.filter(where="NAME LIKE 'Mount%'"))
        assert len(results) == 9
        assert all([x.properties["NAME"].startswith("Mount") for x in results])
        results = list(self.c.filter(where="NAME LIKE '%foo%'"))
        assert len(results) == 0
        results = list(self.c.filter())
        assert len(results) == 67

    def test_filter_bbox_where(self):
        # combined filter criteria
        results = set(self.c.keys(
            bbox=(-120.0, 40.0, -100.0, 50.0), where="NAME LIKE 'Mount%'"))
        assert results == set([0, 2, 5, 13])
        results = set(self.c.keys())
        assert len(results) == 67

    def test_filter_where_error(self):
        for w in ["bad stuff", "NAME=3", "NNAME LIKE 'Mount%'"]:
            with pytest.raises(AttributeFilterError):
                self.c.filter(where=w)

    def test_filter_bbox_where(self):
        # combined filter criteria
        results = set(
            self.c.keys(bbox=(-120.0, 40.0, -100.0, 50.0), where="NAME LIKE 'Mount%'")
        )
        assert results == {0, 2, 5, 13}
        results = set(self.c.keys())
        assert len(results) == 67

    def test_filter_where_error(self):
        for w in ["bad stuff", "NAME=3", "NNAME LIKE 'Mount%'"]:
            with pytest.raises(AttributeFilterError):
                self.c.filter(where=w)


class TestUnsupportedDriver:
    def test_immediate_fail_driver(self, tmpdir):
        schema = {
            "geometry": "Point",
            "properties": {"label": "str", "verit\xe9": "int"},
        }
        with pytest.raises(DriverError):
            fiona.open(str(tmpdir.join("foo")), "w", "Bogus", schema=schema)


@pytest.mark.iconv
class TestGenericWritingTest:
    @pytest.fixture(autouse=True)
    def no_iter_shp(self, tmpdir):
        schema = {
            "geometry": "Point",
            "properties": [("label", "str"), ("verit\xe9", "int")],
        }
        self.c = fiona.open(
            str(tmpdir.join("test-no-iter.shp")),
            "w",
            driver="ESRI Shapefile",
            schema=schema,
            encoding="Windows-1252",
        )
        yield
        self.c.close()

    def test_encoding(self):
        assert self.c.encoding == "Windows-1252"

    def test_no_iter(self):
        with pytest.raises(OSError):
            iter(self.c)

    def test_no_filter(self):
        with pytest.raises(OSError):
            self.c.filter()


class TestPropertiesNumberFormatting:
    @pytest.fixture(autouse=True)
    def shapefile(self, tmpdir):
        self.filename = str(tmpdir.join("properties_number_formatting_test"))

    _records_with_float_property1 = [
        {
            "geometry": {"type": "Point", "coordinates": (0.0, 0.1)},
            "properties": {"property1": 12.22},
        },
        {
            "geometry": {"type": "Point", "coordinates": (0.0, 0.2)},
            "properties": {"property1": 12.88},
        },
    ]

    _records_with_float_property1_as_string = [
        {
            "geometry": {"type": "Point", "coordinates": (0.0, 0.1)},
            "properties": {"property1": "12.22"},
        },
        {
            "geometry": {"type": "Point", "coordinates": (0.0, 0.2)},
            "properties": {"property1": "12.88"},
        },
    ]

    _records_with_invalid_number_property1 = [
        {
            "geometry": {"type": "Point", "coordinates": (0.0, 0.3)},
            "properties": {"property1": "invalid number"},
        }
    ]

    def _write_collection(self, records, schema, driver):
        with fiona.open(
            self.filename,
            "w",
            driver=driver,
            schema=schema,
            crs="epsg:4326",
            encoding="utf-8",
        ) as c:
            c.writerecords([Feature.from_dict(**rec) for rec in records])

    def test_shape_driver_truncates_float_property_to_requested_int_format(self):
        driver = "ESRI Shapefile"
        self._write_collection(
            self._records_with_float_property1,
            {"geometry": "Point", "properties": [("property1", "int")]},
            driver,
        )

        with fiona.open(self.filename, driver=driver, encoding="utf-8") as c:
            assert 2 == len(c)

            rf1, rf2 = list(c)

            assert 12 == rf1.properties["property1"]
            assert 12 == rf2.properties["property1"]

    def test_shape_driver_rounds_float_property_to_requested_digits_number(self):
        driver = "ESRI Shapefile"
        self._write_collection(
            self._records_with_float_property1,
            {"geometry": "Point", "properties": [("property1", "float:15.1")]},
            driver,
        )

        with fiona.open(self.filename, driver=driver, encoding="utf-8") as c:
            assert 2 == len(c)

            rf1, rf2 = list(c)

            assert 12.2 == rf1.properties["property1"]
            assert 12.9 == rf2.properties["property1"]

    def test_string_is_converted_to_number_and_truncated_to_requested_int_by_shape_driver(
        self,
    ):
        driver = "ESRI Shapefile"
        self._write_collection(
            self._records_with_float_property1_as_string,
            {"geometry": "Point", "properties": [("property1", "int")]},
            driver,
        )

        with fiona.open(self.filename, driver=driver, encoding="utf-8") as c:
            assert 2 == len(c)

            rf1, rf2 = list(c)

            assert 12 == rf1.properties["property1"]
            assert 12 == rf2.properties["property1"]

    def test_string_is_converted_to_number_and_rounded_to_requested_digits_number_by_shape_driver(
        self,
    ):
        driver = "ESRI Shapefile"
        self._write_collection(
            self._records_with_float_property1_as_string,
            {"geometry": "Point", "properties": [("property1", "float:15.1")]},
            driver,
        )

        with fiona.open(self.filename, driver=driver, encoding="utf-8") as c:
            assert 2 == len(c)

            rf1, rf2 = list(c)

            assert 12.2 == rf1.properties["property1"]
            assert 12.9 == rf2.properties["property1"]

    def test_invalid_number_is_converted_to_0_and_written_by_shape_driver(self):
        driver = "ESRI Shapefile"
        self._write_collection(
            self._records_with_invalid_number_property1,
            # {'geometry': 'Point', 'properties': [('property1', 'int')]},
            {"geometry": "Point", "properties": [("property1", "float:15.1")]},
            driver,
        )

        with fiona.open(self.filename, driver=driver, encoding="utf-8") as c:
            assert 1 == len(c)

            rf1 = c[0]

            assert 0 == rf1.properties["property1"]

    def test_geojson_driver_truncates_float_property_to_requested_int_format(self):
        driver = "GeoJSON"
        self._write_collection(
            self._records_with_float_property1,
            {"geometry": "Point", "properties": [("property1", "int")]},
            driver,
        )

        with fiona.open(self.filename, driver=driver, encoding="utf-8") as c:
            assert 2 == len(c)

            rf1, rf2 = list(c)

            assert 12 == rf1.properties["property1"]
            assert 12 == rf2.properties["property1"]

    def test_geojson_driver_does_not_round_float_property_to_requested_digits_number(
        self,
    ):
        driver = "GeoJSON"
        self._write_collection(
            self._records_with_float_property1,
            {"geometry": "Point", "properties": [("property1", "float:15.1")]},
            driver,
        )

        with fiona.open(self.filename, driver=driver, encoding="utf-8") as c:
            assert 2 == len(c)

            rf1, rf2 = list(c)

            # ****************************************
            # FLOAT FORMATTING IS NOT RESPECTED...
            assert 12.22 == rf1.properties["property1"]
            assert 12.88 == rf2.properties["property1"]

    def test_string_is_converted_to_number_and_truncated_to_requested_int_by_geojson_driver(
        self,
    ):
        driver = "GeoJSON"
        self._write_collection(
            self._records_with_float_property1_as_string,
            {"geometry": "Point", "properties": [("property1", "int")]},
            driver,
        )

        with fiona.open(self.filename, driver=driver, encoding="utf-8") as c:
            assert 2 == len(c)

            rf1, rf2 = list(c)

            assert 12 == rf1.properties["property1"]
            assert 12 == rf2.properties["property1"]

    def test_string_is_converted_to_number_but_not_rounded_to_requested_digits_number_by_geojson_driver(
        self,
    ):
        driver = "GeoJSON"
        self._write_collection(
            self._records_with_float_property1_as_string,
            {"geometry": "Point", "properties": [("property1", "float:15.1")]},
            driver,
        )

        with fiona.open(self.filename, driver=driver, encoding="utf-8") as c:
            assert 2 == len(c)

            rf1, rf2 = list(c)

            # ****************************************
            # FLOAT FORMATTING IS NOT RESPECTED...
            assert 12.22 == rf1.properties["property1"]
            assert 12.88 == rf2.properties["property1"]

    def test_invalid_number_is_converted_to_0_and_written_by_geojson_driver(self):
        driver = "GeoJSON"
        self._write_collection(
            self._records_with_invalid_number_property1,
            {"geometry": "Point", "properties": [("property1", "float:15.1")]},
            driver,
        )

        with fiona.open(self.filename, driver=driver, encoding="utf-8") as c:
            assert 1 == len(c)

            rf1 = c[0]

            assert 0 == rf1.properties["property1"]


class TestPointWriting:
    @pytest.fixture(autouse=True)
    def shapefile(self, tmpdir):
        self.filename = str(tmpdir.join("point_writing_test.shp"))
        self.sink = fiona.open(
            self.filename,
            "w",
            driver="ESRI Shapefile",
            schema={
                "geometry": "Point",
                "properties": [("title", "str"), ("date", "date")],
            },
            crs="epsg:4326",
            encoding="utf-8",
        )
        yield
        self.sink.close()

    def test_cpg(self, tmpdir):
        """Requires GDAL 1.9"""
        self.sink.close()
        encoding = tmpdir.join("point_writing_test.cpg").read()
        assert encoding == "UTF-8"

    def test_write_one(self):
        assert len(self.sink) == 0
        assert self.sink.bounds == (0.0, 0.0, 0.0, 0.0)
        f = Feature.from_dict(
            **{
                "geometry": {"type": "Point", "coordinates": (0.0, 0.1)},
                "properties": {"title": "point one", "date": "2012-01-29"},
            }
        )
        self.sink.writerecords([f])
        assert len(self.sink) == 1
        assert self.sink.bounds == (0.0, 0.1, 0.0, 0.1)
        self.sink.close()

    def test_write_two(self):
        assert len(self.sink) == 0
        assert self.sink.bounds == (0.0, 0.0, 0.0, 0.0)
        f1 = Feature.from_dict(
            **{
                "geometry": {"type": "Point", "coordinates": (0.0, 0.1)},
                "properties": {"title": "point one", "date": "2012-01-29"},
            }
        )
        f2 = Feature.from_dict(
            **{
                "geometry": {"type": "Point", "coordinates": (0.0, -0.1)},
                "properties": {"title": "point two", "date": "2012-01-29"},
            }
        )
        self.sink.writerecords([f1, f2])
        assert len(self.sink) == 2
        assert self.sink.bounds == (0.0, -0.1, 0.0, 0.1)

    def test_write_one_null_geom(self):
        assert len(self.sink) == 0
        assert self.sink.bounds == (0.0, 0.0, 0.0, 0.0)
        f = Feature.from_dict(
            **{
                "geometry": None,
                "properties": {"title": "point one", "date": "2012-01-29"},
            }
        )
        self.sink.writerecords([f])
        assert len(self.sink) == 1
        assert self.sink.bounds == (0.0, 0.0, 0.0, 0.0)

    def test_validate_record(self):
        fvalid = {
            "geometry": {"type": "Point", "coordinates": (0.0, 0.1)},
            "properties": {"title": "point one", "date": "2012-01-29"},
        }
        finvalid = {
            "geometry": {"type": "Point", "coordinates": (0.0, -0.1)},
            "properties": {"not-a-title": "point two", "date": "2012-01-29"},
        }
        assert self.sink.validate_record(fvalid)
        assert not self.sink.validate_record(finvalid)


class TestLineWriting:
    @pytest.fixture(autouse=True)
    def shapefile(self, tmpdir):
        self.sink = fiona.open(
            str(tmpdir.join("line_writing_test.shp")),
            "w",
            driver="ESRI Shapefile",
            schema={
                "geometry": "LineString",
                "properties": [("title", "str"), ("date", "date")],
            },
            crs={"init": "epsg:4326", "no_defs": True},
        )
        yield
        self.sink.close()

    def test_write_one(self):
        assert len(self.sink) == 0
        assert self.sink.bounds == (0.0, 0.0, 0.0, 0.0)
        f = Feature.from_dict(
            **{
                "geometry": {
                    "type": "LineString",
                    "coordinates": [(0.0, 0.1), (0.0, 0.2)],
                },
                "properties": {"title": "line one", "date": "2012-01-29"},
            }
        )
        self.sink.writerecords([f])
        assert len(self.sink) == 1
        assert self.sink.bounds == (0.0, 0.1, 0.0, 0.2)

    def test_write_two(self):
        assert len(self.sink) == 0
        assert self.sink.bounds == (0.0, 0.0, 0.0, 0.0)
        f1 = Feature.from_dict(
            **{
                "geometry": {
                    "type": "LineString",
                    "coordinates": [(0.0, 0.1), (0.0, 0.2)],
                },
                "properties": {"title": "line one", "date": "2012-01-29"},
            }
        )
        f2 = Feature.from_dict(
            **{
                "geometry": {
                    "type": "MultiLineString",
                    "coordinates": [
                        [(0.0, 0.0), (0.0, -0.1)],
                        [(0.0, -0.1), (0.0, -0.2)],
                    ],
                },
                "properties": {"title": "line two", "date": "2012-01-29"},
            }
        )
        self.sink.writerecords([f1, f2])
        assert len(self.sink) == 2
        assert self.sink.bounds == (0.0, -0.2, 0.0, 0.2)


class TestPointAppend:
    @pytest.fixture(autouse=True)
    def shapefile(self, tmpdir, path_coutwildrnp_shp):
        with fiona.open(path_coutwildrnp_shp, "r") as input:
            output_schema = input.schema
            output_schema["geometry"] = "3D Point"
            with fiona.open(
                str(tmpdir.join("test_append_point.shp")),
                "w",
                crs=None,
                driver="ESRI Shapefile",
                schema=output_schema,
            ) as output:
                for f in input:
                    fnew = Feature(
                        id=f.id,
                        properties=f.properties,
                        geometry=Geometry(
                            type="Point", coordinates=f.geometry.coordinates[0][0]
                        ),
                    )
                    output.write(fnew)

    def test_append_point(self, tmpdir):
        with fiona.open(str(tmpdir.join("test_append_point.shp")), "a") as c:
            assert c.schema["geometry"] == "3D Point"
            c.write(
                Feature.from_dict(
                    **{
                        "geometry": {"type": "Point", "coordinates": (0.0, 45.0)},
                        "properties": {
                            "PERIMETER": 1.0,
                            "FEATURE2": None,
                            "NAME": "Foo",
                            "FEATURE1": None,
                            "URL": "http://example.com",
                            "AGBUR": "BAR",
                            "AREA": 0.0,
                            "STATE_FIPS": 1,
                            "WILDRNP020": 1,
                            "STATE": "XL",
                        },
                    }
                )
            )
            assert len(c) == 68


class TestLineAppend:
    @pytest.fixture(autouse=True)
    def shapefile(self, tmpdir):
        with fiona.open(
            str(tmpdir.join("test_append_line.shp")),
            "w",
            driver="ESRI Shapefile",
            schema={
                "geometry": "MultiLineString",
                "properties": {"title": "str", "date": "date"},
            },
            crs={"init": "epsg:4326", "no_defs": True},
        ) as output:
            f = Feature.from_dict(
                **{
                    "geometry": {
                        "type": "MultiLineString",
                        "coordinates": [[(0.0, 0.1), (0.0, 0.2)]],
                    },
                    "properties": {"title": "line one", "date": "2012-01-29"},
                }
            )
            output.writerecords([f])

    def test_append_line(self, tmpdir):
        with fiona.open(str(tmpdir.join("test_append_line.shp")), "a") as c:
            assert c.schema["geometry"] == "LineString"
            f1 = Feature.from_dict(
                **{
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [(0.0, 0.1), (0.0, 0.2)],
                    },
                    "properties": {"title": "line one", "date": "2012-01-29"},
                }
            )
            f2 = Feature.from_dict(
                **{
                    "geometry": {
                        "type": "MultiLineString",
                        "coordinates": [
                            [(0.0, 0.0), (0.0, -0.1)],
                            [(0.0, -0.1), (0.0, -0.2)],
                        ],
                    },
                    "properties": {"title": "line two", "date": "2012-01-29"},
                }
            )
            c.writerecords([f1, f2])
            assert len(c) == 3
            assert c.bounds == (0.0, -0.2, 0.0, 0.2)


def test_shapefile_field_width(tmpdir):
    name = str(tmpdir.join("textfield.shp"))
    with fiona.open(
        name,
        "w",
        schema={"geometry": "Point", "properties": {"text": "str:254"}},
        driver="ESRI Shapefile",
    ) as c:
        c.write(
            Feature.from_dict(
                **{
                    "geometry": {"type": "Point", "coordinates": (0.0, 45.0)},
                    "properties": {"text": "a" * 254},
                }
            )
        )
    c = fiona.open(name, "r")
    assert c.schema["properties"]["text"] == "str:254"
    f = next(iter(c))
    assert f.properties["text"] == "a" * 254
    c.close()


class TestCollection:
    def test_invalid_mode(self, tmpdir):
        with pytest.raises(ValueError):
            fiona.open(str(tmpdir.join("bogus.shp")), "r+")

    def test_w_args(self, tmpdir):
        with pytest.raises(FionaValueError):
            fiona.open(str(tmpdir.join("test-no-iter.shp")), "w")
        with pytest.raises(FionaValueError):
            fiona.open(str(tmpdir.join("test-no-iter.shp")), "w", "Driver")

    def test_no_path(self):
        with pytest.raises(Exception):
            fiona.open("no-path.shp", "a")

    def test_no_read_conn_str(self):
        with pytest.raises(DriverError):
            fiona.open("PG:dbname=databasename", "r")

    @pytest.mark.skipif(
        sys.platform.startswith("win"), reason="test only for *nix based system"
    )
    def test_no_read_directory(self):
        with pytest.raises(DriverError):
            fiona.open("/dev/null", "r")


def test_date(tmpdir):
    name = str(tmpdir.join("date_test.shp"))
    sink = fiona.open(
        name,
        "w",
        driver="ESRI Shapefile",
        schema={"geometry": "Point", "properties": [("id", "int"), ("date", "date")]},
        crs={"init": "epsg:4326", "no_defs": True},
    )

    recs = [
        Feature.from_dict(
            **{
                "geometry": {"type": "Point", "coordinates": (7.0, 50.0)},
                "properties": {"id": 1, "date": "2013-02-25"},
            }
        ),
        Feature.from_dict(
            **{
                "geometry": {"type": "Point", "coordinates": (7.0, 50.2)},
                "properties": {"id": 1, "date": datetime.date(2014, 2, 3)},
            }
        ),
    ]
    sink.writerecords(recs)
    sink.close()
    assert len(sink) == 2

    with fiona.open(name, "r") as c:
        assert len(c) == 2

        rf1, rf2 = list(c)
        assert rf1.properties["date"] == "2013-02-25"
        assert rf2.properties["date"] == "2014-02-03"


def test_open_kwargs(tmpdir, path_coutwildrnp_shp):
    dstfile = str(tmpdir.join("test.json"))
    with fiona.open(path_coutwildrnp_shp) as src:
        kwds = src.profile
        kwds["driver"] = "GeoJSON"
        kwds["coordinate_precision"] = 2
        with fiona.open(dstfile, "w", **kwds) as dst:
            dst.writerecords(ftr for ftr in src)

    with open(dstfile) as f:
        assert '"coordinates": [ [ [ -111.74, 42.0 ], [ -111.66, 42.0 ]' in f.read(2000)


@pytest.mark.network
def test_collection_http():
    ds = fiona.Collection(
        "https://raw.githubusercontent.com/Toblerity/Fiona/master/tests/data/coutwildrnp.shp",
        vsi="https",
    )
    assert (
        ds.path
        == "/vsicurl/https://raw.githubusercontent.com/Toblerity/Fiona/master/tests/data/coutwildrnp.shp"
    )
    assert len(ds) == 67


@pytest.mark.network
def test_collection_zip_http():
    ds = fiona.Collection(
        "https://raw.githubusercontent.com/Toblerity/Fiona/master/tests/data/coutwildrnp.zip",
        vsi="zip+https",
    )
    assert ds.path == "/vsizip/vsicurl/https://raw.githubusercontent.com/Toblerity/Fiona/master/tests/data/coutwildrnp.zip"
    assert len(ds) == 67


def test_encoding_option_warning(tmpdir, caplog):
    """There is no ENCODING creation option log warning for GeoJSON"""
    with caplog.at_level(logging.WARNING):
        fiona.Collection(
            str(tmpdir.join("test.geojson")),
            "w",
            driver="GeoJSON",
            crs="EPSG:4326",
            schema={"geometry": "Point", "properties": {"foo": "int"}},
            encoding="bogus",
        )
        assert not caplog.text


def test_closed_session_next(gdalenv, path_coutwildrnp_shp):
    """Confirm fix for  issue #687"""
    src = fiona.open(path_coutwildrnp_shp)
    itr = iter(src)
    list(itr)
    src.close()
    with pytest.raises(FionaValueError):
        next(itr)


def test_collection_no_env(path_coutwildrnp_shp):
    """We have no GDAL env left over from open"""
    collection = fiona.open(path_coutwildrnp_shp)
    assert collection
    with pytest.raises(Exception):
        getenv()


def test_collection_env(path_coutwildrnp_shp):
    """We have a GDAL env within collection context"""
    with fiona.open(path_coutwildrnp_shp):
        assert "FIONA_ENV" in getenv()


@pytest.mark.parametrize(
    "driver,filename",
    [("ESRI Shapefile", "test.shp"), ("GeoJSON", "test.json"), ("GPKG", "test.gpkg")],
)
def test_mask_polygon_triangle(tmpdir, driver, filename):
    """Test if mask works for non trivial geometries"""
    schema = {
        "geometry": "Polygon",
        "properties": OrderedDict([("position_i", "int"), ("position_j", "int")]),
    }
    records = [
        Feature.from_dict(
            **{
                "geometry": {
                    "type": "Polygon",
                    "coordinates": (
                        (
                            (float(i), float(j)),
                            (float(i + 1), float(j)),
                            (float(i + 1), float(j + 1)),
                            (float(i), float(j + 1)),
                            (float(i), float(j)),
                        ),
                    ),
                },
                "properties": {"position_i": i, "position_j": j},
            }
        )
        for i in range(10)
        for j in range(10)
    ]
    random.shuffle(records)

    path = str(tmpdir.join(filename))

    with fiona.open(
        path,
        "w",
        driver=driver,
        schema=schema,
    ) as c:
        c.writerecords(records)

    with fiona.open(path) as c:
        items = list(
            c.items(
                mask=Geometry.from_dict(
                    **{
                        "type": "Polygon",
                        "coordinates": (
                            ((2.0, 2.0), (4.0, 4.0), (4.0, 6.0), (2.0, 2.0)),
                        ),
                    }
                )
            )
        )
        assert len(items) == 15


def test_collection__empty_column_name(tmpdir):
    """Based on pull #955"""
    tmpfile = str(tmpdir.join("test_empty.geojson"))
    with pytest.warns(UserWarning, match="Empty field name at index 0"):
        with fiona.open(
            tmpfile,
            "w",
            driver="GeoJSON",
            schema={"geometry": "Point", "properties": {"": "str", "name": "str"}},
        ) as tmp:
            tmp.writerecords(
                [
                    {
                        "geometry": {"type": "Point", "coordinates": [8, 49]},
                        "properties": {"": "", "name": "test"},
                    }
                ]
            )

    with fiona.open(tmpfile) as tmp:
        with pytest.warns(UserWarning, match="Empty field name at index 0"):
            assert tmp.schema == {
                "geometry": "Point",
                "properties": {"": "str", "name": "str"},
            }
        with pytest.warns(UserWarning, match="Empty field name at index 0"):
            next(tmp)


@pytest.mark.parametrize(
    "extension, driver",
    [
        ("shp", "ESRI Shapefile"),
        ("geojson", "GeoJSON"),
        ("json", "GeoJSON"),
        ("gpkg", "GPKG"),
        ("SHP", "ESRI Shapefile"),
    ],
)
def test_driver_detection(tmpdir, extension, driver):
    with fiona.open(
        str(tmpdir.join(f"test.{extension}")),
        "w",
        schema={
            "geometry": "MultiLineString",
            "properties": {"title": "str", "date": "date"},
        },
        crs="EPSG:4326",
    ) as output:
        assert output.driver == driver



"""Tests for ``$ fio info``."""


import json
from pkg_resources import iter_entry_points
import re
import sys

from click.testing import CliRunner
import pytest

from fiona.fio.main import main_group


def test_info_json(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(main_group, ['info', path_coutwildrnp_shp])
    assert result.exit_code == 0
    assert '"count": 67' in result.output
    assert '"crs": "EPSG:4326"' in result.output
    assert '"driver": "ESRI Shapefile"' in result.output
    assert '"name": "coutwildrnp"' in result.output


def test_info_count(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(
        main_group, ['info', '--count', path_coutwildrnp_shp])
    assert result.exit_code == 0
    assert result.output == "67\n"


def test_info_bounds(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(
        main_group, ['info', '--bounds', path_coutwildrnp_shp])
    assert result.exit_code == 0
    assert len(re.findall(r'\d*\.\d*', result.output)) == 4


def test_all_registered():
    """Make sure all the subcommands are actually registered to the main CLI
    group."""
    for ep in iter_entry_points('fiona.fio_commands'):
        assert ep.name in main_group.commands


def _filter_info_warning(lines):
    """$ fio info can issue a RuntimeWarning, but click adds stderr to stdout
    so we have to filter it out before decoding JSON lines."""
    lines = list(filter(lambda x: 'RuntimeWarning' not in x, lines))
    return lines


def test_info_no_count(path_gpx):
    """Make sure we can still get a `$ fio info` report on datasources that do
    not support feature counting, AKA `len(collection)`.
    """
    runner = CliRunner()
    result = runner.invoke(main_group, ['info', path_gpx])
    assert result.exit_code == 0
    lines = _filter_info_warning(result.output.splitlines())
    assert len(lines) == 1, "First line is warning & second is JSON.  No more."
    assert json.loads(lines[0])['count'] is None


def test_info_layer(path_gpx):
    for layer in ('routes', '1'):
        runner = CliRunner()
        result = runner.invoke(main_group, [
            'info',
            path_gpx,
            '--layer', layer])
        assert result.exit_code == 0
        lines = _filter_info_warning(result.output.splitlines())
        assert len(lines) == 1, "1st line is warning & 2nd is JSON - no more."
        assert json.loads(lines[0])['name'] == 'routes'


def test_info_vfs(path_coutwildrnp_zip, path_coutwildrnp_shp):
    runner = CliRunner()
    zip_result = runner.invoke(main_group, [
        'info', f'zip://{path_coutwildrnp_zip}'])
    shp_result = runner.invoke(main_group, [
        'info', path_coutwildrnp_shp])
    assert zip_result.exit_code == shp_result.exit_code == 0
    assert zip_result.output == shp_result.output



import pytest
import fiona
import fiona.drvsupport
import fiona.meta
from fiona.drvsupport import supported_drivers
from fiona.errors import FionaValueError
from .conftest import requires_gdal2, requires_gdal23, requires_gdal31


@requires_gdal31
@pytest.mark.parametrize("driver", supported_drivers)
def test_print_driver_options(driver):
    """ Test fiona.meta.print_driver_options(driver) """
    # do not fail
    fiona.meta.print_driver_options(driver)


@requires_gdal2
def test_metadata_wrong_driver():
    """ Test that FionaValueError is raised for non existing driver"""
    with pytest.raises(FionaValueError):
        fiona.meta.print_driver_options("Not existing driver")


@requires_gdal2
@pytest.mark.parametrize("driver", supported_drivers)
def test_extension(driver):
    """ Test fiona.meta.extension(driver)  """
    # do not fail
    extension = fiona.meta.extension(driver)
    assert extension is None or isinstance(extension, str)


@requires_gdal2
@pytest.mark.parametrize("driver", supported_drivers)
def test_extensions(driver):
    """ Test fiona.meta.extensions(driver) """
    # do not fail
    extensions = fiona.meta.extensions(driver)
    assert extensions is None or isinstance(extensions, list)


@requires_gdal2
@pytest.mark.parametrize("driver", supported_drivers)
def test_supports_vsi(driver):
    """ Test fiona.meta.supports_vsi(driver) """
    # do not fail
    assert fiona.meta.supports_vsi(driver) in (True, False)


@requires_gdal2
@pytest.mark.parametrize("driver", supported_drivers)
def test_supported_field_types(driver):
    """ Test fiona.meta.supported_field_types(driver) """
    # do not fail
    field_types = fiona.meta.supported_field_types(driver)
    assert field_types is None or isinstance(field_types, list)


@requires_gdal23
@pytest.mark.parametrize("driver", supported_drivers)
def test_supported_sub_field_types(driver):
    """ Test fiona.meta.supported_sub_field_types(driver) """
    # do not fail
    sub_field_types = fiona.meta.supported_sub_field_types(driver)
    assert sub_field_types is None or isinstance(sub_field_types, list)



"""Tests of fiona.crs."""

import pytest

from .conftest import requires_gdal33

from fiona import crs
from fiona.env import Env
from fiona.errors import CRSError, FionaDeprecationWarning


def test_proj_keys():
    assert len(crs.all_proj_keys) == 87
    assert 'init' in crs.all_proj_keys
    assert 'proj' in crs.all_proj_keys
    assert 'no_mayo' in crs.all_proj_keys


def test_to_string():
    # Make a string from a mapping with a few bogus items
    val = {
        'proj': 'longlat', 'ellps': 'WGS84', 'datum': 'WGS84',
        'no_defs': True, 'foo': True, 'axis': False, 'belgium': [1, 2]}
    assert crs.CRS.from_user_input(val).to_string() == "EPSG:4326"


def test_to_string_utm():
    # Make a string from a mapping with a few bogus items
    val = {
        'proj': 'utm', 'ellps': 'WGS84', 'zone': 13,
        'no_defs': True, 'foo': True, 'axis': False, 'belgium': [1, 2]}
    assert crs.CRS.from_user_input(val).to_string() == "EPSG:32613"


def test_to_string_epsg():
    val = {'init': 'epsg:4326', 'no_defs': True}
    assert crs.CRS.from_user_input(val).to_string() == "EPSG:4326"


def test_from_epsg():
    val = crs.CRS.from_epsg(4326)
    assert val['init'] == "epsg:4326"


def test_from_epsg_neg():
    with pytest.raises(CRSError):
        crs.CRS.from_epsg(-1)


@pytest.mark.parametrize("invalid_input", [
    "a random string that is invalid",
    ("a", "tuple"),
    "-48567=409 =2095"
])
def test_invalid_crs(invalid_input):
    with pytest.raises(CRSError):
        crs.CRS.from_user_input(invalid_input)


def test_custom_crs():
    class CustomCRS:
        def to_wkt(self):
            return (
                'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",'
                '6378137,298.257223563,AUTHORITY["EPSG","7030"]],'
                'AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,'
                'AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,'
                'AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]'
            )

    assert crs.CRS.from_user_input(CustomCRS()).to_wkt().startswith('GEOGCS["WGS 84"')


def test_crs__version():
    target_crs = (
        'PROJCS["IaRCS_04_Sioux_City-Iowa_Falls_NAD_1983_2011_LCC_US_Feet",'
        'GEOGCS["GCS_NAD_1983_2011",DATUM["D_NAD_1983_2011",'
        'SPHEROID["GRS_1980",6378137.0,298.257222101]],'
        'PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],'
        'PROJECTION["Lambert_Conformal_Conic"],'
        'PARAMETER["False_Easting",14500000.0],'
        'PARAMETER["False_Northing",8600000.0],'
        'PARAMETER["Central_Meridian",-94.83333333333333],'
        'PARAMETER["Standard_Parallel_1",42.53333333333333],'
        'PARAMETER["Standard_Parallel_2",42.53333333333333],'
        'PARAMETER["Scale_Factor",1.000045],'
        'PARAMETER["Latitude_Of_Origin",42.53333333333333],'
        'UNIT["Foot_US",0.3048006096012192]]'
    )
    assert (
        crs.CRS.from_user_input(target_crs)
        .to_wkt(version="WKT2_2018")
        .startswith(
            'PROJCRS["IaRCS_04_Sioux_City-Iowa_Falls_NAD_1983_2011_LCC_US_Feet"'
        )
    )


@requires_gdal33
def test_crs__esri_only_wkt():
    """https://github.com/Toblerity/Fiona/issues/977"""
    target_crs = (
        'PROJCS["IaRCS_04_Sioux_City-Iowa_Falls_NAD_1983_2011_LCC_US_Feet",'
        'GEOGCS["GCS_NAD_1983_2011",DATUM["D_NAD_1983_2011",'
        'SPHEROID["GRS_1980",6378137.0,298.257222101]],'
        'PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],'
        'PROJECTION["Lambert_Conformal_Conic"],'
        'PARAMETER["False_Easting",14500000.0],'
        'PARAMETER["False_Northing",8600000.0],'
        'PARAMETER["Central_Meridian",-94.83333333333333],'
        'PARAMETER["Standard_Parallel_1",42.53333333333333],'
        'PARAMETER["Standard_Parallel_2",42.53333333333333],'
        'PARAMETER["Scale_Factor",1.000045],'
        'PARAMETER["Latitude_Of_Origin",42.53333333333333],'
        'UNIT["Foot_US",0.3048006096012192]]'
    )
    assert (
        crs.CRS.from_user_input(target_crs)
        .to_wkt()
        .startswith(
            (
                'PROJCS["IaRCS_04_Sioux_City-Iowa_Falls_NAD_1983_2011_LCC_US_Feet"',
                'PROJCRS["IaRCS_04_Sioux_City-Iowa_Falls_NAD_1983_2011_LCC_US_Feet"',  # GDAL 3.3+
            )
        )
    )


def test_to_wkt__env_version():
    with Env(OSR_WKT_FORMAT="WKT2_2018"):
        assert crs.CRS.from_string("EPSG:4326").to_wkt().startswith('GEOGCRS["WGS 84",')


def test_to_wkt__invalid_version():
    with pytest.raises(CRSError):
        crs.CRS.from_string("EPSG:4326").to_wkt(version="invalid")


@pytest.mark.parametrize(
    "func, arg",
    [
        (crs.from_epsg, 4326),
        (crs.from_string, "EPSG:4326"),
        (crs.to_string, "EPSG:4326"),
    ],
)
def test_from_func_deprecations(func, arg):
    with pytest.warns(FionaDeprecationWarning):
        _ = func(arg)



"""Fiona's GDAL/AWS environment"""

from functools import wraps, total_ordering
from inspect import getfullargspec
import logging
import os
import re
import threading
import warnings

import attr

from fiona._env import (
    GDALDataFinder,
    GDALEnv,
    PROJDataFinder,
    calc_gdal_version_num,
    get_gdal_config,
    get_gdal_release_name,
    get_gdal_version_num,
    set_gdal_config,
    set_proj_data_search_path,
)
from fiona.errors import EnvError, FionaDeprecationWarning, GDALVersionError
from fiona.session import Session, DummySession


class ThreadEnv(threading.local):
    def __init__(self):
        self._env = None  # Initialises in each thread

        # When the outermost 'fiona.Env()' executes '__enter__' it
        # probes the GDAL environment to see if any of the supplied
        # config options already exist, the assumption being that they
        # were set with 'osgeo.gdal.SetConfigOption()' or possibly
        # 'fiona.env.set_gdal_config()'.  The discovered options are
        # reinstated when the outermost Fiona environment exits.
        # Without this check any environment options that are present in
        # the GDAL environment and are also passed to 'fiona.Env()'
        # will be unset when 'fiona.Env()' tears down, regardless of
        # their value.  For example:
        #
        #   from osgeo import gdal import fiona
        #
        #   gdal.SetConfigOption('key', 'value')
        #   with fiona.Env(key='something'):
        #       pass
        #
        # The config option 'key' would be unset when 'Env()' exits.
        # A more comprehensive solution would also leverage
        # https://trac.osgeo.org/gdal/changeset/37273 but this gets
        # Fiona + older versions of GDAL halfway there.  One major
        # assumption is that environment variables are not set directly
        # with 'osgeo.gdal.SetConfigOption()' OR
        # 'fiona.env.set_gdal_config()' inside of a 'fiona.Env()'.
        self._discovered_options = None


local = ThreadEnv()

log = logging.getLogger(__name__)


class Env:
    """Abstraction for GDAL and AWS configuration

    The GDAL library is stateful: it has a registry of format drivers,
    an error stack, and dozens of configuration options.

    Fiona's approach to working with GDAL is to wrap all the state
    up using a Python context manager (see PEP 343,
    https://www.python.org/dev/peps/pep-0343/). When the context is
    entered GDAL drivers are registered, error handlers are
    configured, and configuration options are set. When the context
    is exited, drivers are removed from the registry and other
    configurations are removed.

    Example:

        with fiona.Env(GDAL_CACHEMAX=512) as env:
            # All drivers are registered, GDAL's raster block cache
            # size is set to 512MB.
            # Commence processing...
            ...
            # End of processing.

        # At this point, configuration options are set to their
        # previous (possible unset) values.

    A boto3 session or boto3 session constructor arguments
    `aws_access_key_id`, `aws_secret_access_key`, `aws_session_token`
    may be passed to Env's constructor. In the latter case, a session
    will be created as soon as needed. AWS credentials are configured
    for GDAL as needed.
    """

    @classmethod
    def default_options(cls):
        """Default configuration options

        Parameters
        ----------
        None

        Returns
        -------
        dict

        """
        return {
            "CHECK_WITH_INVERT_PROJ": True,
            "GTIFF_IMPLICIT_JPEG_OVR": False,
            "FIONA_ENV": True,
        }

    def __init__(
        self,
        session=None,
        aws_unsigned=False,
        profile_name=None,
        session_class=Session.aws_or_dummy,
        **options
    ):
        """Create a new GDAL/AWS environment.
        Note: this class is a context manager. GDAL isn't configured
        until the context is entered via `with fiona.Env():`

        Parameters
        ----------
        session : optional
            A Session object.
        aws_unsigned : bool, optional
            Do not sign cloud requests.
        profile_name : str, optional
            A shared credentials profile name, as per boto3.
        session_class : Session, optional
            A sub-class of Session.
        **options : optional
            A mapping of GDAL configuration options, e.g.,
            `CPL_DEBUG=True, CHECK_WITH_INVERT_PROJ=False`.

        Returns
        -------
        Env

        Notes
        -----
        We raise EnvError if the GDAL config options
        AWS_ACCESS_KEY_ID or AWS_SECRET_ACCESS_KEY are given. AWS
        credentials are handled exclusively by boto3.

        Examples
        --------
        >>> with Env(CPL_DEBUG=True, CPL_CURL_VERBOSE=True):
        ...     with fiona.open("zip+https://example.com/a.zip") as col:
        ...         print(col.profile)

        For access to secured cloud resources, a Fiona Session or a
        foreign session object may be passed to the constructor.

        >>> import boto3
        >>> from fiona.session import AWSSession
        >>> boto3_session = boto3.Session(...)
        >>> with Env(AWSSession(boto3_session)):
        ...     with fiona.open("zip+s3://example/a.zip") as col:
        ...         print(col.profile

        """
        aws_access_key_id = options.pop("aws_access_key_id", None)
        # Warn deprecation in 1.9, remove in 2.0.
        if aws_access_key_id:
            warnings.warn(
                "Passing abstract session keyword arguments is deprecated. "
                "Pass a Fiona AWSSession object instead.",
                FionaDeprecationWarning,
            )

        aws_secret_access_key = options.pop("aws_secret_access_key", None)
        aws_session_token = options.pop("aws_session_token", None)
        region_name = options.pop("region_name", None)

        if not {"AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY"}.isdisjoint(options):
            raise EnvError(
                "GDAL's AWS config options can not be directly set. "
                "AWS credentials are handled exclusively by boto3."
            )

        if session:
            # Passing a session via keyword argument is the canonical
            # way to configure access to secured cloud resources.
            # Warn deprecation in 1.9, remove in 2.0.
            if not isinstance(session, Session):
                warnings.warn(
                    "Passing a boto3 session is deprecated. Pass a Fiona AWSSession object instead.",
                    FionaDeprecationWarning,
                )
                session = Session.aws_or_dummy(session=session)

            self.session = session

        elif aws_access_key_id or profile_name or aws_unsigned:
            self.session = Session.aws_or_dummy(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                region_name=region_name,
                profile_name=profile_name,
                aws_unsigned=aws_unsigned,
            )

        elif {"AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY"}.issubset(os.environ.keys()):
            self.session = Session.from_environ()

        else:
            self.session = DummySession()

        self.options = options.copy()
        self.context_options = {}

    @classmethod
    def from_defaults(cls, *args, **kwargs):
        """Create an environment with default config options

        Parameters
        ----------
        args : optional
            Positional arguments for Env()
        kwargs : optional
            Keyword arguments for Env()

        Returns
        -------
        Env

        Notes
        -----
        The items in kwargs will be overlaid on the default values.

        """
        options = Env.default_options()
        options.update(**kwargs)
        return Env(*args, **options)

    def credentialize(self):
        """Get credentials and configure GDAL

        Note well: this method is a no-op if the GDAL environment
        already has credentials, unless session is not None.

        Returns
        -------
        None

        """
        cred_opts = self.session.get_credential_options()
        self.options.update(**cred_opts)
        setenv(**cred_opts)

    def drivers(self):
        """Return a mapping of registered drivers."""
        return local._env.drivers()

    def _dump_open_datasets(self):
        """Writes descriptions of open datasets to stderr

        For debugging and testing purposes.
        """
        return local._env._dump_open_datasets()

    def __enter__(self):
        if local._env is None:
            self._has_parent_env = False

            # See note directly above where _discovered_options is globally
            # defined.  This MUST happen before calling 'defenv()'.
            local._discovered_options = {}
            # Don't want to reinstate the "RASTERIO_ENV" option.
            probe_env = {k for k in self.options.keys() if k != "RASTERIO_ENV"}
            for key in probe_env:
                val = get_gdal_config(key, normalize=False)
                if val is not None:
                    local._discovered_options[key] = val

            defenv(**self.options)
            self.context_options = {}
        else:
            self._has_parent_env = True
            self.context_options = getenv()
            setenv(**self.options)

        self.credentialize()
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        delenv()
        if self._has_parent_env:
            defenv()
            setenv(**self.context_options)
        else:
            # See note directly above where _discovered_options is globally
            # defined.
            while local._discovered_options:
                key, val = local._discovered_options.popitem()
                set_gdal_config(key, val, normalize=False)

            local._discovered_options = None


def defenv(**options):
    """Create a default environment if necessary."""
    if not local._env:
        local._env = GDALEnv()
        local._env.update_config_options(**options)

    local._env.start()


def getenv():
    """Get a mapping of current options."""
    if not local._env:
        raise EnvError("No GDAL environment exists")
    else:
        return local._env.options.copy()


def hasenv():
    return bool(local._env)


def setenv(**options):
    """Set options in the existing environment."""
    if not local._env:
        raise EnvError("No GDAL environment exists")
    else:
        local._env.update_config_options(**options)


def hascreds():
    warnings.warn("Please use Env.session.hascreds() instead", FionaDeprecationWarning)
    return local._env is not None and all(
        key in local._env.get_config_options()
        for key in ["AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY"]
    )


def delenv():
    """Delete options in the existing environment."""
    if not local._env:
        raise EnvError("No GDAL environment exists")
    else:
        local._env.clear_config_options()

    local._env.stop()
    local._env = None


class NullContextManager:
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


def env_ctx_if_needed():
    """Return an Env if one does not exist

    Returns
    -------
    Env or a do-nothing context manager

    """
    if local._env:
        return NullContextManager()
    else:
        return Env.from_defaults()


def ensure_env(f):
    """A decorator that ensures an env exists before a function
    calls any GDAL C functions.

    Parameters
    ----------
    f : function
        A function.

    Returns
    -------
    A function wrapper.

    Notes
    -----
    If there is already an existing environment, the wrapper does
    nothing and immediately calls f with the given arguments.

    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if local._env:
            return f(*args, **kwargs)
        else:
            with Env.from_defaults():
                return f(*args, **kwargs)

    return wrapper


def ensure_env_with_credentials(f):
    """Ensures a config environment exists and has credentials.

    Parameters
    ----------
    f : function
        A function.

    Returns
    -------
    A function wrapper.

    Notes
    -----
    The function wrapper checks the first argument of f and
    credentializes the environment if the first argument is a URI with
    scheme "s3".

    If there is already an existing environment, the wrapper does
    nothing and immediately calls f with the given arguments.

    """

    @wraps(f)
    def wrapper(*args, **kwds):
        if local._env:
            env_ctor = Env
        else:
            env_ctor = Env.from_defaults

        fp_arg = kwds.get("fp", None) or args[0]

        if isinstance(fp_arg, str):
            session_cls = Session.cls_from_path(fp_arg)

            if local._env and session_cls.hascreds(getenv()):
                session_cls = DummySession

            session = session_cls()

        else:
            session = DummySession()

        with env_ctor(session=session):
            return f(*args, **kwds)

    return wrapper


@attr.s(slots=True)
@total_ordering
class GDALVersion:
    """Convenience class for obtaining GDAL major and minor version
    components and comparing between versions.  This is highly
    simplistic and assumes a very normal numbering scheme for versions
    and ignores everything except the major and minor components.

    """

    major = attr.ib(default=0, validator=attr.validators.instance_of(int))
    minor = attr.ib(default=0, validator=attr.validators.instance_of(int))

    def __eq__(self, other):
        return (self.major, self.minor) == tuple(other.major, other.minor)

    def __lt__(self, other):
        return (self.major, self.minor) < tuple(other.major, other.minor)

    def __repr__(self):
        return f"GDALVersion(major={self.major}, minor={self.minor})"

    def __str__(self):
        return f"{self.major}.{self.minor}"

    @classmethod
    def parse(cls, input):
        """
        Parses input tuple or string to GDALVersion. If input is a GDALVersion
        instance, it is returned.

        Parameters
        ----------
        input: tuple of (major, minor), string, or instance of GDALVersion

        Returns
        -------
        GDALVersion instance

        """
        if isinstance(input, cls):
            return input
        if isinstance(input, tuple):
            return cls(*input)
        elif isinstance(input, str):
            # Extract major and minor version components.
            # alpha, beta, rc suffixes ignored
            match = re.search(r"^\d+\.\d+", input)
            if not match:
                raise ValueError(
                    "value does not appear to be a valid GDAL version "
                    "number: {}".format(input)
                )
            major, minor = (int(c) for c in match.group().split("."))
            return cls(major=major, minor=minor)

        raise TypeError("GDALVersion can only be parsed from a string or tuple")

    @classmethod
    def runtime(cls):
        """Return GDALVersion of current GDAL runtime"""
        return cls.parse(get_gdal_release_name())

    def at_least(self, other):
        other = self.__class__.parse(other)
        return self >= other


def require_gdal_version(
    version, param=None, values=None, is_max_version=False, reason=""
):
    """A decorator that ensures the called function or parameters are supported
    by the runtime version of GDAL.  Raises GDALVersionError if conditions
    are not met.

    Examples:
    \b
        @require_gdal_version('2.2')
        def some_func():

    calling `some_func` with a runtime version of GDAL that is < 2.2 raises a
    GDALVersionErorr.

    \b
        @require_gdal_version('2.2', param='foo')
        def some_func(foo='bar'):

    calling `some_func` with parameter `foo` of any value on GDAL < 2.2 raises
    a GDALVersionError.

    \b
        @require_gdal_version('2.2', param='foo', values=('bar',))
        def some_func(foo=None):

    calling `some_func` with parameter `foo` and value `bar` on GDAL < 2.2
    raises a GDALVersionError.


    Parameters
    ------------
    version: tuple, string, or GDALVersion
    param: string (optional, default: None)
        If `values` are absent, then all use of this parameter with a value
        other than default value requires at least GDAL `version`.
    values: tuple, list, or set (optional, default: None)
        contains values that require at least GDAL `version`.  `param`
        is required for `values`.
    is_max_version: bool (optional, default: False)
        if `True` indicates that the version provided is the maximum version
        allowed, instead of requiring at least that version.
    reason: string (optional: default: '')
        custom error message presented to user in addition to message about
        GDAL version.  Use this to provide an explanation of what changed
        if necessary context to the user.

    Returns
    ---------
    wrapped function

    """
    if values is not None:
        if param is None:
            raise ValueError("require_gdal_version: param must be provided with values")

        if not isinstance(values, (tuple, list, set)):
            raise ValueError(
                "require_gdal_version: values must be a tuple, list, or set"
            )

    version = GDALVersion.parse(version)
    runtime = GDALVersion.runtime()
    inequality = ">=" if runtime < version else "<="
    reason = f"\n{reason}" if reason else reason

    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwds):
            if (runtime < version and not is_max_version) or (
                is_max_version and runtime > version
            ):

                if param is None:
                    raise GDALVersionError(
                        "GDAL version must be {} {}{}".format(
                            inequality, str(version), reason
                        )
                    )

                # normalize args and kwds to dict
                argspec = getfullargspec(f)
                full_kwds = kwds.copy()

                if argspec.args:
                    full_kwds.update(dict(zip(argspec.args[: len(args)], args)))

                if argspec.defaults:
                    defaults = dict(
                        zip(reversed(argspec.args), reversed(argspec.defaults))
                    )
                else:
                    defaults = {}

                if param in full_kwds:
                    if values is None:
                        if param not in defaults or (
                            full_kwds[param] != defaults[param]
                        ):
                            raise GDALVersionError(
                                'usage of parameter "{}" requires '
                                "GDAL {} {}{}".format(
                                    param, inequality, version, reason
                                )
                            )

                    elif full_kwds[param] in values:
                        raise GDALVersionError(
                            'parameter "{}={}" requires '
                            "GDAL {} {}{}".format(
                                param, full_kwds[param], inequality, version, reason
                            )
                        )

            return f(*args, **kwds)

        return wrapper

    return decorator


# Patch the environment if needed, such as in the installed wheel case.

if "GDAL_DATA" not in os.environ:

    path = GDALDataFinder().search_wheel()

    if path:
        log.debug("GDAL data found in package: path=%r.", path)
        set_gdal_config("GDAL_DATA", path)

    # See https://github.com/mapbox/rasterio/issues/1631.
    elif GDALDataFinder().find_file("header.dxf"):
        log.debug("GDAL data files are available at built-in paths.")

    else:
        path = GDALDataFinder().search()

        if path:
            set_gdal_config("GDAL_DATA", path)
            log.debug("GDAL data found in other locations: path=%r.", path)

if 'PROJ_DATA' in os.environ:
    # PROJ 9.1+
    path = os.environ["PROJ_DATA"]
    set_proj_data_search_path(path)

elif "PROJ_LIB" in os.environ:
    # PROJ < 9.1
    path = os.environ["PROJ_LIB"]
    set_proj_data_search_path(path)

elif PROJDataFinder().search_wheel():
    path = PROJDataFinder().search_wheel()
    log.debug("PROJ data found in package: path=%r.", path)
    set_proj_data_search_path(path)

# See https://github.com/mapbox/rasterio/issues/1631.
elif PROJDataFinder().has_data():
    log.debug("PROJ data files are available at built-in paths.")

else:
    path = PROJDataFinder().search()

    if path:
        log.debug("PROJ data found in other locations: path=%r.", path)
        set_proj_data_search_path(path)



"""
Tests related to the validation of feature geometry types against the schema.
"""

import fiona
import pytest

from fiona.errors import GeometryTypeValidationError, UnsupportedGeometryTypeError
from fiona.model import Feature


@pytest.fixture
def filename_shp(tmpdir):
    return str(tmpdir.join("example.shp"))


@pytest.fixture
def filename_json(tmpdir):
    return str(tmpdir.join("example.json"))


properties = {"name": "str"}
PROPERTIES = {"name": "example"}
POINT = {"type": "Point", "coordinates": (1.0, 2.0)}
LINESTRING = {"type": "LineString", "coordinates": [(1.0, 2.0), (3.0, 4.0)]}
POLYGON = {"type": "Polygon", "coordinates": [[(0.0, 0.0), (1.0, 1.0), (0.0, 0.1)]]}
MULTILINESTRING = {
    "type": "MultiLineString",
    "coordinates": [[(0.0, 0.0), (1.0, 1.0)], [(1.0, 2.0), (3.0, 4.0)]],
}
GEOMETRYCOLLECTION = {
    "type": "GeometryCollection",
    "geometries": [POINT, LINESTRING, POLYGON],
}
INVALID = {"type": "InvalidType", "coordinates": (42.0, 43.0)}
POINT_3D = {"type": "Point", "coordinates": (1.0, 2.0, 3.0)}


def write_point(collection):
    feature = Feature.from_dict(**{"geometry": POINT, "properties": PROPERTIES})
    collection.write(feature)


def write_linestring(collection):
    feature = Feature.from_dict(**{"geometry": LINESTRING, "properties": PROPERTIES})
    collection.write(feature)


def write_polygon(collection):
    feature = Feature.from_dict(**{"geometry": POLYGON, "properties": PROPERTIES})
    collection.write(feature)


def write_invalid(collection):
    feature = Feature.from_dict(**{"geometry": INVALID, "properties": PROPERTIES})
    collection.write(feature)


def write_multilinestring(collection):
    feature = Feature.from_dict(
        **{"geometry": MULTILINESTRING, "properties": PROPERTIES}
    )
    collection.write(feature)


def write_point_3d(collection):
    feature = Feature.from_dict(**{"geometry": POINT_3D, "properties": PROPERTIES})
    collection.write(feature)


def write_geometrycollection(collection):
    feature = Feature.from_dict(
        **{"geometry": GEOMETRYCOLLECTION, "properties": PROPERTIES}
    )
    collection.write(feature)


def write_null(collection):
    feature = Feature.from_dict(**{"geometry": None, "properties": PROPERTIES})
    collection.write(feature)


def test_point(filename_shp):
    schema = {"geometry": "Point", "properties": properties}
    with fiona.open(
        filename_shp, "w", driver="ESRI Shapefile", schema=schema
    ) as collection:
        write_point(collection)
        write_point_3d(collection)
        write_null(collection)

        with pytest.raises(GeometryTypeValidationError):
            write_linestring(collection)

        with pytest.raises(GeometryTypeValidationError):
            write_invalid(collection)


def test_multi_type(filename_json):
    schema = {"geometry": ("Point", "LineString"), "properties": properties}
    with fiona.open(filename_json, "w", driver="GeoJSON", schema=schema) as collection:
        write_point(collection)
        write_linestring(collection)
        write_null(collection)

        with pytest.raises(GeometryTypeValidationError):
            write_polygon(collection)

        with pytest.raises(GeometryTypeValidationError):
            write_invalid(collection)


def test_unknown(filename_json):
    """Reading and writing layers with "Unknown" (i.e. any) geometry type"""
    # write a layer with a mixture of geometry types
    schema = {"geometry": "Unknown", "properties": properties}
    with fiona.open(filename_json, "w", driver="GeoJSON", schema=schema) as collection:
        write_point(collection)
        write_linestring(collection)
        write_polygon(collection)
        write_geometrycollection(collection)
        write_null(collection)

        with pytest.raises(GeometryTypeValidationError):
            write_invalid(collection)

    # copy the features to a new layer, reusing the layers metadata
    with fiona.open(filename_json, "r", driver="GeoJSON") as src:
        filename_dst = filename_json.replace(".json", "_v2.json")
        assert src.schema["geometry"] == "Unknown"
        with fiona.open(filename_dst, "w", **src.meta) as dst:
            dst.writerecords(src)


def test_any(filename_json):
    schema = {"geometry": "Any", "properties": properties}
    with fiona.open(filename_json, "w", driver="GeoJSON", schema=schema) as collection:
        write_point(collection)
        write_linestring(collection)
        write_polygon(collection)
        write_geometrycollection(collection)
        write_null(collection)

        with pytest.raises(GeometryTypeValidationError):
            write_invalid(collection)


def test_broken(filename_json):
    schema = {"geometry": "NOT_VALID", "properties": properties}
    with pytest.raises(UnsupportedGeometryTypeError):
        with fiona.open(filename_json, "w", driver="GeoJSON", schema=schema):
            pass


def test_broken_list(filename_json):
    schema = {
        "geometry": ("Point", "LineString", "NOT_VALID"),
        "properties": properties,
    }
    with pytest.raises(UnsupportedGeometryTypeError):
        collection = fiona.open(filename_json, "w", driver="GeoJSON", schema=schema)


def test_invalid_schema(filename_shp):
    """Features match schema but geometries not supported by driver"""
    schema = {"geometry": ("Point", "LineString"), "properties": properties}
    with fiona.open(
        filename_shp, "w", driver="ESRI Shapefile", schema=schema
    ) as collection:
        write_linestring(collection)

        with pytest.raises(RuntimeError):
            # ESRI Shapefile can only store a single geometry type
            write_point(collection)


def test_esri_multi_geom(filename_shp):
    """ESRI Shapefile doesn't differentiate between LineString/MultiLineString"""
    schema = {"geometry": "LineString", "properties": properties}
    with fiona.open(
        filename_shp, "w", driver="ESRI Shapefile", schema=schema
    ) as collection:
        write_linestring(collection)
        write_multilinestring(collection)

        with pytest.raises(GeometryTypeValidationError):
            write_point(collection)


def test_3d_schema_ignored(filename_json):
    schema = {"geometry": "3D Point", "properties": properties}
    with fiona.open(filename_json, "w", driver="GeoJSON", schema=schema) as collection:
        write_point(collection)
        write_point_3d(collection)


def test_geometrycollection_schema(filename_json):
    schema = {"geometry": "GeometryCollection", "properties": properties}
    with fiona.open(filename_json, "w", driver="GeoJSON", schema=schema) as collection:
        write_geometrycollection(collection)


def test_none_schema(filename_json):
    schema = {"geometry": None, "properties": properties}
    with fiona.open(filename_json, "w", driver="GeoJSON", schema=schema) as collection:
        write_null(collection)

        with pytest.raises(GeometryTypeValidationError):
            write_point(collection)
        with pytest.raises(GeometryTypeValidationError):
            write_linestring(collection)



from collections import OrderedDict
import glob
import os
import tempfile

import fiona
from fiona.model import Feature

from .conftest import get_temp_filename, requires_gdal2


def test_gml_format_option(tmp_path):
    """Test GML dataset creation option FORMAT (see gh-968)"""

    schema = {"geometry": "Point", "properties": OrderedDict([("position", "int")])}
    records = [
        Feature.from_dict(
            **{
                "geometry": {"type": "Point", "coordinates": (0.0, float(i))},
                "properties": {"position": i},
            }
        )
        for i in range(10)
    ]

    fpath = tmp_path.joinpath(get_temp_filename("GML"))

    with fiona.open(fpath, "w", driver="GML", schema=schema, FORMAT="GML3") as out:
        out.writerecords(records)

    xsd_path = list(tmp_path.glob("*.xsd"))[0]

    with open(xsd_path) as f:
        xsd = f.read()
        assert "http://schemas.opengis.net/gml/3.1.1" in xsd



import json
import os.path
import tempfile

import fiona
from fiona import prop_type, prop_width
from fiona.model import Feature
from fiona.rfc3339 import FionaDateType


def test_width_str():
    assert prop_width("str:254") == 254
    assert prop_width("str") == 80


def test_width_other():
    assert prop_width("int") == None
    assert prop_width("float") == None
    assert prop_width("date") == None


def test_types():
    assert prop_type("str:254") == str
    assert prop_type("str") == str
    assert isinstance(0, prop_type("int"))
    assert isinstance(0.0, prop_type("float"))
    assert prop_type("date") == FionaDateType


def test_read_json_object_properties():
    """JSON object properties are properly serialized"""
    data = """
{
  "type": "FeatureCollection",
  "features": [
    {
      "geometry": {
        "type": "Polygon",
        "coordinates": [
          [
            [
              87.33588,
              43.53139
            ],
            [
              87.33588,
              45.66894
            ],
            [
              90.27542,
              45.66894
            ],
            [
              90.27542,
              43.53139
            ],
            [
              87.33588,
              43.53139
            ]
          ]
        ]
      },
      "type": "Feature",
      "properties": {
        "upperLeftCoordinate": {
          "latitude": 45.66894,
          "longitude": 87.91166
        },
        "tricky": "{gotcha"
      }
    }
  ]
}
"""
    tmpdir = tempfile.mkdtemp()
    filename = os.path.join(tmpdir, "test.json")

    with open(filename, "w") as f:
        f.write(data)

    with fiona.open(filename) as src:
        ftr = next(iter(src))
        props = ftr["properties"]
        assert props["upperLeftCoordinate"]["latitude"] == 45.66894
        assert props["upperLeftCoordinate"]["longitude"] == 87.91166
        assert props["tricky"] == "{gotcha"


def test_write_json_object_properties():
    """Python object properties are properly serialized"""
    data = """
{
  "type": "FeatureCollection",
  "features": [
    {
      "geometry": {
        "type": "Polygon",
        "coordinates": [
          [
            [
              87.33588,
              43.53139
            ],
            [
              87.33588,
              45.66894
            ],
            [
              90.27542,
              45.66894
            ],
            [
              90.27542,
              43.53139
            ],
            [
              87.33588,
              43.53139
            ]
          ]
        ]
      },
      "type": "Feature",
      "properties": {
        "upperLeftCoordinate": {
          "latitude": 45.66894,
          "longitude": 87.91166
        },
        "tricky": "{gotcha"
      }
    }
  ]
}
"""
    data = Feature.from_dict(**json.loads(data)["features"][0])
    tmpdir = tempfile.mkdtemp()
    filename = os.path.join(tmpdir, "test.json")
    with fiona.open(
        filename,
        "w",
        driver="GeoJSON",
        schema={
            "geometry": "Polygon",
            "properties": {"upperLeftCoordinate": "str", "tricky": "str"},
        },
    ) as dst:
        dst.write(data)

    with fiona.open(filename) as src:
        ftr = next(iter(src))
        props = ftr["properties"]
        assert props["upperLeftCoordinate"]["latitude"] == 45.66894
        assert props["upperLeftCoordinate"]["longitude"] == 87.91166
        assert props["tricky"] == "{gotcha"


def test_json_prop_decode_non_geojson_driver():
    feature = Feature.from_dict(
        **{
            "type": "Feature",
            "properties": {
                "ulc": {"latitude": 45.66894, "longitude": 87.91166},
                "tricky": "{gotcha",
            },
            "geometry": {"type": "Point", "coordinates": [10, 15]},
        }
    )

    meta = {
        "crs": "EPSG:4326",
        "driver": "ESRI Shapefile",
        "schema": {
            "geometry": "Point",
            "properties": {"ulc": "str:255", "tricky": "str:255"},
        },
    }

    tmpdir = tempfile.mkdtemp()
    filename = os.path.join(tmpdir, "test.json")
    with fiona.open(filename, "w", **meta) as dst:
        dst.write(feature)

    with fiona.open(filename) as src:
        actual = next(iter(src))

    assert isinstance(actual["properties"]["ulc"], str)
    a = json.loads(actual["properties"]["ulc"])
    e = json.loads(actual["properties"]["ulc"])
    assert e == a
    assert actual["properties"]["tricky"].startswith("{")



from collections import defaultdict
import logging
import os
import pytest
from random import uniform, randint

import fiona
from fiona.model import Feature
import fiona.ogrext

from tests.conftest import requires_gdal2

has_gpkg = "GPKG" in fiona.supported_drivers.keys()


def create_records(count):
    for n in range(count):
        record = {
            "geometry": {
                "type": "Point",
                "coordinates": [uniform(-180, 180), uniform(-90, 90)],
            },
            "properties": {"value": randint(0, 1000)},
        }
        yield Feature.from_dict(**record)


class DebugHandler(logging.Handler):
    def __init__(self, pattern):
        logging.Handler.__init__(self)
        self.pattern = pattern
        self.history = defaultdict(lambda: 0)

    def emit(self, record):
        if self.pattern in record.msg:
            self.history[record.msg] += 1


log = logging.getLogger()


@requires_gdal2
@pytest.mark.skipif(not has_gpkg, reason="Requires geopackage driver")
class TestTransaction:
    def setup_method(self):
        self.handler = DebugHandler(pattern="transaction")
        self.handler.setLevel(logging.DEBUG)
        log.setLevel(logging.DEBUG)
        log.addHandler(self.handler)

    def teardown_method(self):
        log.removeHandler(self.handler)

    def test_transaction(self, tmpdir):
        """
        Test transaction start/commit is called the expected number of times,
        and that the default transaction size can be overloaded. The test uses
        a custom logging handler to listen for the debug messages produced
        when the transaction is started/committed.
        """
        num_records = 250
        transaction_size = 100

        assert fiona.ogrext.DEFAULT_TRANSACTION_SIZE == 20000
        fiona.ogrext.DEFAULT_TRANSACTION_SIZE = transaction_size
        assert fiona.ogrext.DEFAULT_TRANSACTION_SIZE == transaction_size

        path = str(tmpdir.join("output.gpkg"))

        schema = {"geometry": "Point", "properties": {"value": "int"}}

        with fiona.open(path, "w", driver="GPKG", schema=schema) as dst:
            dst.writerecords(create_records(num_records))

        assert self.handler.history["Starting transaction (initial)"] == 1
        assert (
            self.handler.history["Starting transaction (intermediate)"]
            == num_records // transaction_size
        )
        assert (
            self.handler.history["Committing transaction (intermediate)"]
            == num_records // transaction_size
        )
        assert self.handler.history["Committing transaction (final)"] == 1

        with fiona.open(path, "r") as src:
            assert len(src) == num_records



"""Encoding tests"""

from glob import glob
import os
import shutil

import pytest

import fiona

from .conftest import requires_gdal2


@pytest.fixture(scope='function')
def gre_shp_cp1252(tmpdir):
    """A tempdir containing copies of gre.* files, .cpg set to cp1252

    The shapefile attributes are in fact utf-8 encoded.
    """
    test_files = glob(os.path.join(os.path.dirname(__file__), 'data/gre.*'))
    tmpdir = tmpdir.mkdir('data')
    for filename in test_files:
        shutil.copy(filename, str(tmpdir))
    tmpdir.join('gre.cpg').write('CP1252')
    yield tmpdir.join('gre.shp')


@requires_gdal2
def test_broken_encoding(gre_shp_cp1252):
    """Reading as cp1252 mis-encodes a Russian name"""
    with fiona.open(str(gre_shp_cp1252)) as src:
        assert src.session._get_internal_encoding() == 'utf-8'
        feat = next(iter(src))
        assert feat['properties']['name_ru'] != 'Гренада'


@requires_gdal2
def test_cpg_encoding(gre_shp_cp1252):
    """Reads a Russian name"""
    gre_shp_cp1252.join('../gre.cpg').write('UTF-8')
    with fiona.open(str(gre_shp_cp1252)) as src:
        assert src.session._get_internal_encoding() == 'utf-8'
        feat = next(iter(src))
        assert feat['properties']['name_ru'] == 'Гренада'


@requires_gdal2
def test_override_encoding(gre_shp_cp1252):
    """utf-8 override succeeds"""
    with fiona.open(str(gre_shp_cp1252), encoding='utf-8') as src:
        assert src.session._get_internal_encoding() == 'utf-8'
        assert next(iter(src))['properties']['name_ru'] == 'Гренада'



"""Tests for Fiona's RFC 3339 support."""


import re

import pytest

from fiona.rfc3339 import parse_date, parse_datetime, parse_time
from fiona.rfc3339 import group_accessor, pattern_date


class TestDateParse:

    def test_yyyymmdd(self):
        assert parse_date("2012-01-29") == (2012, 1, 29, 0, 0, 0, 0.0, None)

    def test_error(self):
        with pytest.raises(ValueError):
            parse_date("xxx")


class TestTimeParse:

    def test_hhmmss(self):
        assert parse_time("10:11:12") == (0, 0, 0, 10, 11, 12, 0.0, None)

    def test_hhmm(self):
        assert parse_time("10:11") == (0, 0, 0, 10, 11, 0, 0.0, None)

    def test_hhmmssff(self):
        assert parse_time("10:11:12.42") == (0, 0, 0, 10, 11, 12, 0.42*1000000, None)

    def test_hhmmssz(self):
        assert parse_time("10:11:12Z") == (0, 0, 0, 10, 11, 12, 0.0, None)

    def test_hhmmssoff(self):
        assert parse_time("10:11:12-01:30") == (0, 0, 0, 10, 11, 12, 0.0, -90)

    def test_hhmmssoff2(self):
        assert parse_time("10:11:12+01:30") == (0, 0, 0, 10, 11, 12, 0.0, 90)

    def test_error(self):
        with pytest.raises(ValueError):
            parse_time("xxx")


class TestDatetimeParse:

    def test_yyyymmdd(self):
        assert (
            parse_datetime("2012-01-29T10:11:12") ==
            (2012, 1, 29, 10, 11, 12, 0.0, None))

    def test_yyyymmddTZ(self):
        assert (
            parse_datetime("2012-01-29T10:11:12+01:30") ==
            (2012, 1, 29, 10, 11, 12, 0.0, 90))

    def test_yyyymmddTZ2(self):
        assert (
            parse_datetime("2012-01-29T10:11:12-01:30") ==
            (2012, 1, 29, 10, 11, 12, 0.0, -90))

    def test_error(self):
        with pytest.raises(ValueError):
            parse_datetime("xxx")


def test_group_accessor_indexerror():
    match = re.search(pattern_date, '2012-01-29')
    g = group_accessor(match)
    assert g.group(-1) == 0
    assert g.group(6) == 0



"""fio-collect"""

from functools import partial
import json
import logging

import click
import cligj

from fiona.fio import helpers, options, with_context_env
from fiona.model import Geometry, ObjectEncoder
from fiona.transform import transform_geom


@click.command(short_help="Collect a sequence of features.")
@cligj.precision_opt
@cligj.indent_opt
@cligj.compact_opt
@click.option(
    "--record-buffered/--no-record-buffered",
    default=False,
    help="Economical buffering of writes at record, not collection "
    "(default), level.",
)
@click.option(
    "--ignore-errors/--no-ignore-errors",
    default=False,
    help="log errors but do not stop serialization.",
)
@options.src_crs_opt
@click.option(
    "--with-ld-context/--without-ld-context",
    default=False,
    help="add a JSON-LD context to JSON output.",
)
@click.option(
    "--add-ld-context-item",
    multiple=True,
    help="map a term to a URI and add it to the output's JSON LD " "context.",
)
@click.option(
    "--parse/--no-parse",
    default=True,
    help="load and dump the geojson feature (default is True)",
)
@click.pass_context
@with_context_env
def collect(
    ctx,
    precision,
    indent,
    compact,
    record_buffered,
    ignore_errors,
    src_crs,
    with_ld_context,
    add_ld_context_item,
    parse,
):
    """Make a GeoJSON feature collection from a sequence of GeoJSON
    features and print it."""
    logger = logging.getLogger(__name__)
    stdin = click.get_text_stream("stdin")
    sink = click.get_text_stream("stdout")

    dump_kwds = {"sort_keys": True}
    if indent:
        dump_kwds["indent"] = indent
    if compact:
        dump_kwds["separators"] = (",", ":")
    item_sep = compact and "," or ", "

    if src_crs:
        if not parse:
            raise click.UsageError("Can't specify --src-crs with --no-parse")
        transformer = partial(
            transform_geom,
            src_crs,
            "EPSG:4326",
            antimeridian_cutting=True,
            precision=precision,
        )
    else:

        def transformer(x):
            return x

    first_line = next(stdin)

    # If parsing geojson
    if parse:
        # If input is RS-delimited JSON sequence.
        if first_line.startswith("\x1e"):

            def feature_text_gen():
                buffer = first_line.strip("\x1e")
                for line in stdin:
                    if line.startswith("\x1e"):
                        if buffer:
                            feat = json.loads(buffer)
                            feat["geometry"] = transformer(
                                Geometry.from_dict(**feat["geometry"])
                            )
                            yield json.dumps(feat, cls=ObjectEncoder, **dump_kwds)
                        buffer = line.strip("\x1e")
                    else:
                        buffer += line
                else:
                    feat = json.loads(buffer)
                    feat["geometry"] = transformer(
                        Geometry.from_dict(**feat["geometry"])
                    )
                    yield json.dumps(feat, cls=ObjectEncoder, **dump_kwds)

        else:

            def feature_text_gen():
                feat = json.loads(first_line)
                feat["geometry"] = transformer(Geometry.from_dict(**feat["geometry"]))
                yield json.dumps(feat, cls=ObjectEncoder, **dump_kwds)

                for line in stdin:
                    feat = json.loads(line)
                    feat["geometry"] = transformer(
                        Geometry.from_dict(**feat["geometry"])
                    )
                    yield json.dumps(feat, cls=ObjectEncoder, **dump_kwds)

    # If *not* parsing geojson
    else:
        # If input is RS-delimited JSON sequence.
        if first_line.startswith("\x1e"):

            def feature_text_gen():
                buffer = first_line.strip("\x1e")
                for line in stdin:
                    if line.startswith("\x1e"):
                        if buffer:
                            yield buffer
                        buffer = line.strip("\x1e")
                    else:
                        buffer += line
                else:
                    yield buffer

        else:

            def feature_text_gen():
                yield first_line
                yield from stdin

    source = feature_text_gen()

    if record_buffered:
        # Buffer GeoJSON data at the feature level for smaller
        # memory footprint.
        indented = bool(indent)
        rec_indent = "\n" + " " * (2 * (indent or 0))

        collection = {"type": "FeatureCollection", "features": []}
        if with_ld_context:
            collection["@context"] = helpers.make_ld_context(add_ld_context_item)

        head, tail = json.dumps(collection, cls=ObjectEncoder, **dump_kwds).split("[]")

        sink.write(head)
        sink.write("[")

        # Try the first record.
        try:
            i, first = 0, next(source)
            if with_ld_context:
                first = helpers.id_record(first)
            if indented:
                sink.write(rec_indent)
            sink.write(first.replace("\n", rec_indent))
        except StopIteration:
            pass
        except Exception as exc:
            # Ignoring errors is *not* the default.
            if ignore_errors:
                logger.error(
                    "failed to serialize file record %d (%s), " "continuing", i, exc
                )
            else:
                # Log error and close up the GeoJSON, leaving it
                # more or less valid no matter what happens above.
                logger.critical(
                    "failed to serialize file record %d (%s), " "quiting", i, exc
                )
                sink.write("]")
                sink.write(tail)
                if indented:
                    sink.write("\n")
                raise

        # Because trailing commas aren't valid in JSON arrays
        # we'll write the item separator before each of the
        # remaining features.
        for i, rec in enumerate(source, 1):
            try:
                if with_ld_context:
                    rec = helpers.id_record(rec)
                if indented:
                    sink.write(rec_indent)
                sink.write(item_sep)
                sink.write(rec.replace("\n", rec_indent))
            except Exception as exc:
                if ignore_errors:
                    logger.error(
                        "failed to serialize file record %d (%s), " "continuing",
                        i,
                        exc,
                    )
                else:
                    logger.critical(
                        "failed to serialize file record %d (%s), " "quiting",
                        i,
                        exc,
                    )
                    sink.write("]")
                    sink.write(tail)
                    if indented:
                        sink.write("\n")
                    raise

        # Close up the GeoJSON after writing all features.
        sink.write("]")
        sink.write(tail)
        if indented:
            sink.write("\n")

    else:
        # Buffer GeoJSON data at the collection level. The default.
        collection = {"type": "FeatureCollection", "features": []}
        if with_ld_context:
            collection["@context"] = helpers.make_ld_context(add_ld_context_item)

        head, tail = json.dumps(collection, cls=ObjectEncoder, **dump_kwds).split("[]")
        sink.write(head)
        sink.write("[")
        sink.write(",".join(source))
        sink.write("]")
        sink.write(tail)
        sink.write("\n")



"""Binary BLOB field testing."""

import pytest
import struct
from collections import OrderedDict

import fiona
from fiona.model import Feature

from .conftest import requires_gpkg


@requires_gpkg
def test_binary_field(tmpdir):
    meta = {
        "driver": "GPKG",
        "schema": {
            "geometry": "Point",
            "properties": OrderedDict(
                [
                    ("name", "str"),
                    ("data", "bytes"),
                ]
            ),
        },
    }

    # create some binary data
    input_data = struct.pack("256B", *range(256))

    # write the binary data to a BLOB field
    filename = str(tmpdir.join("binary_test.gpkg"))
    with fiona.open(filename, "w", **meta) as dst:
        feature = Feature.from_dict(
            **{
                "geometry": {"type": "Point", "coordinates": ((0, 0))},
                "properties": {
                    "name": "test",
                    "data": input_data,
                },
            }
        )
        dst.write(feature)

    # read the data back and check consistency
    with fiona.open(filename, "r") as src:
        feature = next(iter(src))
        assert feature.properties["name"] == "test"
        output_data = feature.properties["data"]
        assert output_data == input_data
